/*
 * jidct_bin_l3.c
 *
 * binDCT from Loeffler's fact: Var. 3: 16 shifts, 34 adds.
 * Use floor operation for all right-shifting.
 */

/*
 ************************************************
 *
 * $Log: jidct_bin_l3.c,v $
 * Revision 1.1  2000/07/23 15:38:08  jliang
 * Initial revision
 *
 * Revision 1.1  2000/06/26 01:06:44  jliang
 * Initial revision
 *
 *
 *
 ************************************************
 */

/*
**********************************************************************
*
* Modification History:
* Date       Programmer   Description
* --------   ----------   --------------------------------------------
*
***********************************************************************
*/

#include "sim.h"

extern int lossless_codec;

int idct_bin_l3(int *coef_block, int *output_buf)
{
  INT32 tmp0, tmp1, tmp2, tmp3,tmp4,tmp5,tmp6,tmp7;
  INT32 tmp10, tmp11, tmp12, tmp13;
  INT32 z1;
  JCOEFPTR inptr;
  int * wsptr;
  JSAMPROW outptr;
  int ctr, i;
  int workspace[DCTSIZE2];	/* buffers data between passes */
  int dcval;

//Case 1: lossless binDCT not required. All scaling caused by butterfly
//are performed at the last stage.
if (!lossless_codec){

  /* Pass 1: process columns from input, store into work array. */
  /* Note results are scaled up by sqrt(8) compared to a true IDCT; */
  /* furthermore, we scale the results by 2**PASS1_BITS. */

  inptr = coef_block;
  wsptr = workspace;
  for (ctr = DCTSIZE; ctr > 0; ctr--) {


    // If one side uses this short-cut technique, the other also has to use it,
    // othersie the reconstructed results are different.
    //For example, if one row is 71 0 0 0 0 0 0 0, then by this fast mtd,
    //the resul is all 35s, but if we go through the inverse bindct, some 
    //value will be 35, and some will be 36.
    //Jie Liang, 10/21/2000

    //
    // check all 0 AC 
    // since the shortcut in the 2nd pass will descale by 32,
    // here it's only descaled by 2. overall descale is still 64.
    if (inptr[DCTSIZE*1] == 0 && inptr[DCTSIZE*2] == 0 &&
	inptr[DCTSIZE*3] == 0 && inptr[DCTSIZE*4] == 0 &&
	inptr[DCTSIZE*5] == 0 && inptr[DCTSIZE*6] == 0 &&
	inptr[DCTSIZE*7] == 0) {
      dcval = inptr[DCTSIZE*0] >> 1;
      
      wsptr[DCTSIZE*0] = dcval;
      wsptr[DCTSIZE*1] = dcval;
      wsptr[DCTSIZE*2] = dcval;
      wsptr[DCTSIZE*3] = dcval;
      wsptr[DCTSIZE*4] = dcval;
      wsptr[DCTSIZE*5] = dcval;
      wsptr[DCTSIZE*6] = dcval;
      wsptr[DCTSIZE*7] = dcval;
      
      inptr++;			
      wsptr++;
      continue;
    }

    tmp0 = inptr[DCTSIZE*0];
    tmp1 = inptr[DCTSIZE*4];
    tmp2 = inptr[DCTSIZE*6];
    tmp3 = inptr[DCTSIZE*2];
    tmp4 = inptr[DCTSIZE*7];
    tmp5 = inptr[DCTSIZE*3];
    tmp6 = inptr[DCTSIZE*5];  //different from Chen.
    tmp7 = inptr[DCTSIZE*1];

    /* X[0] and X[4] */
    tmp11 = ((tmp0 ) >> 1) - tmp1;
    tmp10 = tmp0 - tmp11;
	
    /* X[6] and X[2]: 3/8, 7/16 */
    tmp13 = tmp3 + (((tmp2 << 1) + tmp2 ) >> 3);
    tmp12 = (((tmp13 << 3) - tmp13 ) >> 4) - tmp2;

    tmp0 = tmp10 + tmp13;
    tmp3 = tmp10 - tmp13;
    tmp1 = tmp11 + tmp12;
    tmp2 = tmp11 - tmp12;

    /* last lift between X[7] and X[1]: */
    tmp4 = ((tmp7 ) >> 1) - tmp4;
    tmp7 = tmp7 - tmp4;

    //2 intermediate butterflies
    tmp10 = tmp4 + tmp6;
    tmp11 = tmp7 - tmp5;
    tmp12 = tmp4 - tmp6;
    tmp13 = tmp5 + tmp7;

    //pi/16: 3/32, -3/16, 1/8
    tmp12 = (((tmp11 << 1) + tmp11 ) >> 5) + tmp12;
    tmp5  = tmp11 - (((tmp12 << 1) + tmp12 ) >> 4);
    tmp6  = ((tmp5 ) >> 3) + tmp12;

    //3pi/16: 5/16, -9/16, 1/4
    tmp13 = (((tmp10 << 2) + tmp10 ) >> 4) + tmp13;
    tmp4  = tmp10 - (((tmp13 << 3) + tmp13 ) >> 4);
    tmp7  = ((tmp4 ) >> 2) + tmp13;

    /* last stage: butterfly */
    wsptr[DCTSIZE*0] = (tmp0 + tmp7);
    wsptr[DCTSIZE*7] = (tmp0 - tmp7);
    wsptr[DCTSIZE*1] = (tmp1 + tmp6);
    wsptr[DCTSIZE*6] = (tmp1 - tmp6);
    wsptr[DCTSIZE*2] = (tmp2 + tmp5);
    wsptr[DCTSIZE*5] = (tmp2 - tmp5);
    wsptr[DCTSIZE*3] = (tmp3 + tmp4);
    wsptr[DCTSIZE*4] = (tmp3 - tmp4);
    
    inptr++;			/* advance pointers to next column */
    wsptr++;
  }
  
  /* Pass 2: process rows from work array, store into output array. */
  /* Note that we must descale the results by a factor of 8 == 2**3, */
  /* and also undo the PASS1_BITS scaling. */

  //fprintf(stderr, "\nAfter inverse DCT:\n");

  wsptr = workspace;
  outptr = output_buf;
  for (ctr = 0; ctr < DCTSIZE; ctr++) {
    //    outptr = output_buf[ctr] + output_col;

   //shortcut
    if (wsptr[1] == 0 && wsptr[2] == 0 && wsptr[3] == 0 && wsptr[4] == 0 &&
	wsptr[5] == 0 && wsptr[6] == 0 && wsptr[7] == 0) {

      //if all AC are 0, the IDCT will all equal to 1/2 DC, so downscale by 2,
      //After that, apply the downscale of 16 caused by butterflies, so total downscale = 32.
      dcval = DESCALE(wsptr[0], 5);
      outptr[0] = dcval;
      outptr[1] = dcval;
      outptr[2] = dcval;
      outptr[3] = dcval;
      outptr[4] = dcval;
      outptr[5] = dcval;
      outptr[6] = dcval;
      outptr[7] = dcval;

      wsptr += DCTSIZE;		
      outptr += DCTSIZE;     //H.263
      continue;
    }

    /* X[0] and X[4] */
    tmp11 = ((wsptr[0] ) >> 1) - wsptr[4];
    tmp10 = wsptr[0] - tmp11;
	
    /* X[6] and X[2]: 3/8, 7/16 */
    tmp13 = wsptr[2] + (((wsptr[6] << 1) + wsptr[6] ) >> 3);
    tmp12 = (((tmp13 << 3) - tmp13 ) >> 4) - wsptr[6];
    
    tmp0 = tmp10 + tmp13;
    tmp3 = tmp10 - tmp13;
    tmp1 = tmp11 + tmp12;
    tmp2 = tmp11 - tmp12;

    /* last lift between X[7] and X[1]: */
    tmp4 = ((wsptr[1] ) >> 1) - wsptr[7];
    tmp7 = wsptr[1] - tmp4;

    //different from Chen's factorization.
    tmp5 = wsptr[3];
    tmp6 = wsptr[5];

    //2 intermediate butterflies
    tmp10 = tmp4 + tmp6;
    tmp11 = tmp7 - tmp5;
    tmp12 = tmp4 - tmp6;
    tmp13 = tmp5 + tmp7;

    //pi/16: 3/32, -3/16, 1/8
    tmp12 = (((tmp11 << 1) + tmp11 ) >> 5) + tmp12;
    tmp5  = tmp11 - (((tmp12 << 1) + tmp12 ) >> 4);
    tmp6  = ((tmp5 ) >> 3) + tmp12;

    //3pi/16: 5/16, -9/16, 1/4
    tmp13 = (((tmp10 << 2) + tmp10 ) >> 4) + tmp13;
    tmp4  = tmp10 - (((tmp13 << 3) + tmp13 ) >> 4);
    tmp7  = ((tmp4 ) >> 2) + tmp13;

    /* last stage: butterfly */

    /* Final output stage: scale down by a factor of 8 and range-limit */
    tmp10=(tmp0 + tmp7);
    tmp11=(tmp0 - tmp7);
    outptr[0] = DESCALE(tmp10, 4 );
    outptr[7] = DESCALE(tmp11, 4 );

    tmp10=(tmp1 + tmp6);
    tmp11=(tmp1 - tmp6);
    outptr[1] = DESCALE(tmp10, 4);
    outptr[6] = DESCALE(tmp11, 4);

    tmp10=(tmp2 + tmp5);
    tmp11=(tmp2 - tmp5);
    outptr[2] = DESCALE(tmp10, 4);
    outptr[5] = DESCALE(tmp11, 4) ;

    tmp10=(tmp3 + tmp4);
    tmp11=(tmp3 - tmp4);
    outptr[3] = DESCALE(tmp10, 4);
    outptr[4] = DESCALE(tmp11, 4);

    wsptr += DCTSIZE;		/* advance pointer to next row */

    outptr += DCTSIZE;

/*******************/
/* Jie: test code */
	/*	for (tmp0 = 0; tmp0 < 8; tmp0 ++) {
	  fprintf(stderr, "%10d", outptr[tmp0]);
	}
	fprintf(stderr, "\n");
	*/
  }

/****************************************************************************/
   
} else {

/****************************************************************************/
  //Case 2: lossless binDCT: descale by 2 immediately after inverse butterfly.


  /* Pass 1: process columns from input, store into work array. */
  /* Note results are scaled up by sqrt(8) compared to a true IDCT; */
  /* furthermore, we scale the results by 2**PASS1_BITS. */

  inptr = coef_block;
  wsptr = workspace;
  for (ctr = DCTSIZE; ctr > 0; ctr--) {
	
    tmp0 = inptr[DCTSIZE*0];
    tmp1 = inptr[DCTSIZE*4];
    tmp2 = inptr[DCTSIZE*6];
    tmp3 = inptr[DCTSIZE*2];
    tmp4 = inptr[DCTSIZE*7];
    tmp5 = inptr[DCTSIZE*3];  //diff from Chen
    tmp6 = inptr[DCTSIZE*5];  // Diff from Chen
    tmp7 = inptr[DCTSIZE*1];

	//tmp10 = tmp0 - tmp11;
	tmp10 = tmp0 + ((tmp1 ) >> 1);
	tmp11 = tmp10 - tmp1;
	
	/* X[6] and X[2]: 3/8, 7/16 */
	tmp13 = tmp3 + (((tmp2 << 1) + tmp2 ) >> 3);
	tmp12 = (((tmp13 << 3) - tmp13 ) >> 4) - tmp2;

	//lossless binDCT: use new nutterflies.
	//tmp3 = ((tmp10 ) >> 1) - tmp13;
	//tmp0 = tmp10 - tmp3;
	tmp0 = tmp10 + ((tmp13 ) >> 1);
	tmp3 = tmp0 - tmp13;

	//tmp2 = ((tmp11 ) >> 1) - tmp12;
	//tmp1 = tmp11 - tmp2;
	tmp1 = tmp11 + ((tmp12 ) >> 1);
	tmp2 = tmp1 - tmp12;

	// odd part

	// X[7] and X[1]: butterfly
	tmp7 = tmp7 + ((tmp4 ) >> 1);
	tmp4 = tmp7 - tmp4;

	//intermediate bf
	tmp10 = tmp4 + (( tmp6 ) >> 1);
	tmp12 = tmp10 - tmp6;

	tmp13 = ((tmp5 ) >> 1) + tmp7;
	tmp11 = tmp13 - tmp5;

	//pi/16: 3/32, -3/16, 1/8
	tmp12 = (((tmp11 << 1) + tmp11 ) >> 5) + tmp12;
	tmp5  = tmp11 - (((tmp12 << 1) + tmp12 ) >> 4);
	tmp6  = ((tmp5 ) >> 3) + tmp12;

	//3pi/16: 5/16, -9/16, 1/4
	tmp13 = (((tmp10 << 2) + tmp10 ) >> 4) + tmp13;
	tmp4  = tmp10 - (((tmp13 << 3) + tmp13 ) >> 4);
	tmp7  = ((tmp4 ) >> 2) + tmp13;

 	/* last stage: butterfly */
	wsptr[DCTSIZE*0] = tmp0 + ((tmp7 ) >> 1);
    wsptr[DCTSIZE*7] = wsptr[DCTSIZE*0] - tmp7 ;

    wsptr[DCTSIZE*1] = tmp1 + ((tmp6 ) >> 1);
    wsptr[DCTSIZE*6] = wsptr[DCTSIZE*1] - tmp6 ;

    wsptr[DCTSIZE*2] = tmp2 + ((tmp5 ) >> 1);
    wsptr[DCTSIZE*5] = wsptr[DCTSIZE*2] - tmp5 ;

    wsptr[DCTSIZE*3] = tmp3 + ((tmp4 ) >> 1);
    wsptr[DCTSIZE*4] = wsptr[DCTSIZE*3] - tmp4 ;
 
    inptr++;			/* advance pointers to next column */
    wsptr++;
  }
  
  /* Pass 2: process rows from work array, store into output array. */
  /* Note that we must descale the results by a factor of 8 == 2**3, */
  /* and also undo the PASS1_BITS scaling. */

  //fprintf(stderr, "\nAfter inverse DCT:\n");

  wsptr = workspace;
  outptr = output_buf;
  for (ctr = 0; ctr < DCTSIZE; ctr++) {
    //    outptr = output_buf[ctr] + output_col;
    
    /* Even part: reverse the even part of the forward DCT. */
    /* The rotator is sqrt(2)*c(-6). */

	/* X[0] and X[4] */
	//tmp11 = ((wsptr[0] ) >> 1) - wsptr[4];
	//tmp10 = wsptr[0] - tmp11;
	tmp10 = wsptr[0] + ((wsptr[4] ) >> 1);
	tmp11 = tmp10 - wsptr[4];
	
	/* X[6] and X[2]:3/8, 7/16 */
	tmp13 = wsptr[2] + (((wsptr[6] << 1) + wsptr[6] ) >> 3);
	tmp12 = (((tmp13 << 3) - tmp13 ) >> 4) - wsptr[6];

	//lossless binDCT: use new nutterflies.
	//	tmp3 = ((tmp10 ) >> 1) - tmp13;
	//	tmp0 = tmp10 - tmp3;
	tmp0 = tmp10 + ((tmp13 ) >> 1);
	tmp3 = tmp0 - tmp13;

	//	tmp2 = ((tmp11 ) >> 1) - tmp12;
	//	tmp1 = tmp11 - tmp2;
	tmp1 = tmp11 + ((tmp12 ) >> 1);
	tmp2 = tmp1 - tmp12;

	// odd part

	// X[7] and X[1]: butterfly
	tmp7 = wsptr[1] + ((wsptr[7] ) >> 1);
	tmp4 = tmp7 - wsptr[7];

	tmp5 = wsptr[3];
	tmp6 = wsptr[5];

	//intermediate bf
	tmp10 = tmp4 + (( tmp6 ) >> 1);
	tmp12 = tmp10 - tmp6;

	tmp13 = ((tmp5 ) >> 1) + tmp7;
	tmp11 = tmp13 - tmp5;

	//pi/16: 3/32, -3/16, 1/8
	tmp12 = (((tmp11 << 1) + tmp11 ) >> 5) + tmp12;
	tmp5  = tmp11 - (((tmp12 << 1) + tmp12 ) >> 4);
	tmp6  = ((tmp5 ) >> 3) + tmp12;

	//3pi/16: 5/16, -9/16, 1/4
	tmp13 = (((tmp10 << 2) + tmp10 ) >> 4) + tmp13;
	tmp4  = tmp10 - (((tmp13 << 3) + tmp13 ) >> 4);
	tmp7  = ((tmp4 ) >> 2) + tmp13;

/* last stage: butterfly */

    /* Final output stage: scale down by a factor of 8 and range-limit */
    tmp10=tmp0 + ((tmp7 ) >> 1);
    tmp11=tmp10 - tmp7;
    //	outptr[0] = range_limit[tmp10 & RANGE_MASK];
    //	outptr[7] = range_limit[tmp11 & RANGE_MASK];
    outptr[0] = tmp10;
    outptr[7] = tmp11;

    tmp10=tmp1 + ((tmp6 ) >> 1);
    tmp11=(tmp10 - tmp6);
    //    outptr[1] = range_limit[tmp10 & RANGE_MASK];
    //    outptr[6] = range_limit[tmp11 & RANGE_MASK];
    outptr[1] =  tmp10;
    outptr[6] =  tmp11;

    tmp10=tmp2 + ((tmp5 ) >> 1);
    tmp11=(tmp10 - tmp5);
    //	outptr[2] = range_limit[tmp10 & RANGE_MASK];
    //	outptr[5] = range_limit[tmp11 & RANGE_MASK];
    outptr[2] = tmp10;
    outptr[5] = tmp11;

    tmp10=tmp3 + ((tmp4 ) >> 1);
    tmp11=(tmp10 - tmp4);
    //	outptr[3] = range_limit[tmp10 & RANGE_MASK];
    //	outptr[4] = range_limit[tmp11 & RANGE_MASK];   
    outptr[3] =  tmp10;
    outptr[4] =  tmp11;

	/* 	for (z1 = 0; z1 < 8; z1 ++) {
	  fprintf(stderr, "%6d", outptr[z1]-128);
	}
	fprintf(stderr, "\n");*/
     
    wsptr += DCTSIZE;		/* advance pointer to next row */
    outptr += DCTSIZE;

/*******************/
/* Jie: test code */
	/*	for (tmp0 = 0; tmp0 < 8; tmp0 ++) {
	  fprintf(stderr, "%10d", outptr[tmp0]);
	}
	fprintf(stderr, "\n");
	*/
  }

}

 return(0);

}

